% Least-squares Hermite Radial Basis Functions Implicits
% Copyright (c) 2013 Harlen Costa Batagelo, harlen.batagelo@ufabc.edu.br

function lshrbf_pou

clc;
close all;

if isdeployed
  old = pwd;
  cd(ctfroot);
end

pou_margin_sides = zeros(6,1);
eval_options = zeros(5,1);
approx_options = zeros(2,1);
[h, max_div, num_cen, pou_overlap, pou_margin_factor, pou_margin_sides, eval_options, greedy_options, approx_options] = lshrbf_pou_gui();
delete(h);
use_least_squares = approx_options(1);
if max_div == -1
  return
end

if eval_options(2) == 0
  eval_options(5) = 0;
end

[filein, dirin] = uigetfile('*.ply', 'Select Input Point Cloud File');
if filein == 0
  return
end

%[tmpa, name, tmpc] = fileparts(filein); % R2009a
[~, name, ~] = fileparts(filein); % R2011+
[fileout, dirout] = uiputfile(sprintf('%s', name, '.dat'), 'Save Output Coefficient File As');
if fileout == 0
  return
end

if eval_options(5)
  [eval_fileout, eval_dirout] = uiputfile(sprintf('%s', name, '.obj'), 'Save Triangle Mesh File As');
  if eval_fileout == 0
    eval_options(5) = 0;
  end
end

if isdeployed
  cd(old);
end

stream = RandStream('mt19937ar','seed', 0);
RandStream.setDefaultStream(stream);

filename = fullfile(dirin, filein);
grid_size = eval_options(4);
remove_false_sheets = approx_options(2);

use_random = false;

eye3 = eye(3);

filename_centers = 'twist_75_ls.samples';
use_filename_centers = false;
if use_filename_centers
  fid = fopen(filename_centers);
  num_loaded_centers = fscanf(fid, '%u', 1); % Number of points
  tmp = fscanf(fid, '%f'); % Point (xyz) and normal (xyz)
  for i=1:num_loaded_centers
    loaded_centers(i,1:6) = tmp((i-1)*6+1:(i-1)*6+6);
  end
  fclose(fid);
end

max_gdga_con = 100000000;

dmsz = [8 8 8];

res = '-r300';

num_cen_global = 1;

% Read point cloud

h = waitbar(0,'Reading file...',...
            'Name', 'Loading point cloud. Please wait...');
fid = fopen(filename);
%[tmpa,tmp,data] = ply_read(filename,'tri'); % r2009
[~,tmp,data] = ply_read(filename,'tri'); % r2011+
num_pt = size(tmp,2);
fclose(fid);
waitbar(0.5, h, 'Removing null normals...');  

mVP_orig = zeros(num_pt, 3);
mVN_orig = zeros(num_pt, 3);
pmin = [1e10 1e10 1e10];
pmax = -pmin;
centroid = zeros(1,3);

% Remove vertices with null normals
mVP_orig2 = zeros(num_pt, 3);
mVN_orig2 = zeros(num_pt, 3);
j = 0;
for i=1:num_pt
    N = [data.vertex.nx(i), data.vertex.ny(i), data.vertex.nz(i)];
    if norm(N) > 0
      j = j + 1;      
      mVP_orig2(j,1:3) = [data.vertex.x(i), data.vertex.y(i), data.vertex.z(i)];
      mVN_orig2(j,1:3) = N;
    end
end
clear N; clear mVP_orig; clear mVN_orig;
mVP_orig = mVP_orig2(1:j,1:3);
mVN_orig = mVN_orig2(1:j,1:3);
num_pt = j;
clear mVP_orig2; clear mVN_orig2;

for i=1:num_pt
  mag = norm(mVN_orig(i,1:3));
  if mag > 0
    mVN_orig(i,1:3) = mVN_orig(i,1:3) ./ mag;
  else
    disp('Error. This model contains null normals.');
    return;
  end
  
  centroid = centroid + mVP_orig(i,1:3);
  if mVP_orig(i,1) < pmin(1); pmin(1) = mVP_orig(i,1); end
  if mVP_orig(i,2) < pmin(2); pmin(2) = mVP_orig(i,2); end
  if mVP_orig(i,3) < pmin(3); pmin(3) = mVP_orig(i,3); end
    
  if mVP_orig(i,1) > pmax(1); pmax(1) = mVP_orig(i,1); end
  if mVP_orig(i,2) > pmax(2); pmax(2) = mVP_orig(i,2); end
  if mVP_orig(i,3) > pmax(3); pmax(3) = mVP_orig(i,3); end
end

% Center and normalize size
psize = pmax - pmin;
centroid = psize / 2;
for i=1:num_pt
  mVP_orig(i,1:3) = ((mVP_orig(i,1:3) - pmin - centroid) / max(psize));
  mVP_orig(i,1:3) = mVP_orig(i,1:3) * 10;
end

waitbar(1, h, sprintf('Success! %d points loaded.', num_pt));
pause(1.5);
delete(h);
clear tmp; clear num_pt;

min_pou = min(mVP_orig,[],1);
max_pou = max(mVP_orig,[],1);
pou_size = max_pou - min_pou;

% Add margin
if pou_margin_sides(1)
  max_pou(1) = max_pou(1) + pou_size(1) * pou_margin_factor;
end
if pou_margin_sides(2)
  min_pou(1) = min_pou(1) - pou_size(1) * pou_margin_factor;
end
if pou_margin_sides(3)
  max_pou(2) = max_pou(2) + pou_size(2) * pou_margin_factor;
end
if pou_margin_sides(4)
  min_pou(2) = min_pou(2) - pou_size(2) * pou_margin_factor;
end
if pou_margin_sides(5)
  max_pou(3) = max_pou(3) + pou_size(3) * pou_margin_factor;
end
if pou_margin_sides(6)
  min_pou(3) = min_pou(3) - pou_size(3) * pou_margin_factor;
end

pou_size(1) = max_pou(1) - min_pou(1);
pou_size(2) = max_pou(2) - min_pou(2);
pou_size(3) = max_pou(3) - min_pou(3);

pou_center = (min_pou + max_pou) / 2;

if max(pou_size) == pou_size(1)
  M = grid_size;
  N = grid_size / (pou_size(1)/pou_size(2));
  O = grid_size / (pou_size(1)/pou_size(3));
  pou_div = [max_div, max_div/(pou_size(1)/pou_size(2)), max_div/(pou_size(1)/pou_size(2))];
elseif max(pou_size) == pou_size(2)
  M = grid_size / (pou_size(2)/pou_size(1));
  N = grid_size;
  O = grid_size / (pou_size(2)/pou_size(3));
  pou_div = [max_div/(pou_size(2)/pou_size(1)), max_div, max_div/(pou_size(2)/pou_size(3))];
else
  M = grid_size / (pou_size(3)/pou_size(1));
  N = grid_size / (pou_size(3)/pou_size(2));
  O = grid_size;
  pou_div = [max_div/(pou_size(3)/pou_size(1)), max_div/(pou_size(3)/pou_size(2)), max_div];
end
pou_div = ceil(pou_div);
pou_step = pou_size./pou_div;

M = round(M);
N = round(N);
O = round(O);

lsv_scale = greedy_options(1)/norm(max_pou - min_pou);

clc;
cells = {};
cells.number = pou_div(1) * pou_div(2) * pou_div(3);
cells.S = zeros(cells.number, 3);
cells.T = zeros(cells.number, 3);
l=1;
overlap = (pou_overlap-1)*(pou_step)/2;
iint = 1;

h = waitbar(0,'Computing partition domain boxes (0%)',...
            'Name', 'Preprocessing 1/2. Please wait...',...
            'CreateCancelBtn',...
            'setappdata(gcbf,''canceling'',1)');
setappdata(h,'canceling',0);

tot = pou_div(1) * pou_div(2) * pou_div(3);
for i=0:pou_step(1):pou_size(1)-pou_step(1)
  if getappdata(h,'canceling')
    delete(h);
    return;
  end  
  
  jint = 1;
  for j=0:pou_step(2):pou_size(2)-pou_step(2)
    kint = 1;
    for k=0:pou_step(3):pou_size(3)-pou_step(3)
      S = [i j k] + min_pou - overlap;
      T = [i j k] + min_pou + pou_step + overlap;
      
      cells.S(l,:) = S;
      cells.T(l,:) = T;
      cells.idx(l,:) = [iint jint kint];
      l = l + 1;
      kint = kint + 1;
    end
    jint = jint + 1;
  end
  iint = iint + 1;
  
  prog = l/tot;
  waitbar(prog, h, sprintf('Computing partition domain boxes (%5.1f%%)', prog*100));  
end
clear tot;
delete(h);

% Pre-discretize domains
big_size = max(cells.T,[],1)-min(cells.S,[],1);
dfac = ([M N O]-1).*(big_size./pou_size);
minS = min(cells.S,[],1);
half_cell = (([.5 .5 .5] ./ dfac) .* big_size);
for l=1:cells.number
  S = cells.S(l,:);
  T = cells.T(l,:);
  
  dS = floor(((S-minS)./big_size) .* dfac);
  dT = ceil(((T-minS)./big_size) .* dfac);
  cells.S(l,:) = ((dS ./ dfac) .* big_size) + minS;
  cells.T(l,:) = ((dT ./ dfac) .* big_size) + minS;
  
  cells.dS(l,:) = dS + 1;
  cells.dT(l,:) = dT + 1;
end

% Determine number of points inside current cell
h = waitbar(0,'Computing number of points inside each cell (0%)',...
            'Name', 'Preprocessing 2/2. Please wait...',...
            'CreateCancelBtn',...
            'setappdata(gcbf,''canceling'',1)');
setappdata(h,'canceling',0)

for cur_cell=1:cells.number
  cells.mVP_num{cur_cell} = 0;
end

% For each point...
szmVPorig = size(mVP_orig,1);
if szmVPorig/10 > cells.number
  for cur_cell=1:cells.number
    if getappdata(h,'canceling')
      delete(h);
      return;
    end
    
    S = cells.S(cur_cell,1:3);
    T = cells.T(cur_cell,1:3);
    for l=1:szmVPorig
      if mVP_orig(l,1) >= S(1) && mVP_orig(l,2) >= S(2) && mVP_orig(l,3) >= S(3) && mVP_orig(l,1) <= T(1) && mVP_orig(l,2) <= T(2) && mVP_orig(l,3) <= T(3)
        cells.mVP_num{cur_cell} = cells.mVP_num{cur_cell} + 1;
        idx = cells.mVP_num{cur_cell};
        cells.mVP{cur_cell}(idx,:) = mVP_orig(l,:);
        cells.mVN{cur_cell}(idx,:) = mVN_orig(l,:);
      end
    end
    
    prog = (l/szmVPorig);
    waitbar(prog, h, sprintf('Computing number of points inside each cell (%5.1f%%)', prog*100));        
  end  
else
  for l=1:szmVPorig
    if getappdata(h,'canceling')
      delete(h);
      return;
    end  

    pt = mVP_orig(l,:);

    % For each neighbour
    cur_cell_int = round((((pt - min_pou) ./ pou_size) .* (pou_div-1)) + 1);
    for i=-1:1
      for j=-1:1
        for k=-1:1
          nei_int = cur_cell_int + [i j k];
          if nei_int(1) < 1 || nei_int(2) < 1 || nei_int(3) < 1 || nei_int(1) > pou_div(1) || nei_int(2) > pou_div(2) || nei_int(3) > pou_div(3)
            continue % This cell is out of bounds
          end
          cur_cell = ((nei_int(1)-1) * pou_div(3) * pou_div(2)) + ((nei_int(2)-1) * pou_div(3)) + nei_int(3);
          S = cells.S(cur_cell,1:3);
          T = cells.T(cur_cell,1:3);
          if pt(1) >= S(1) && pt(1) <= T(1) && pt(2) >= S(2) && pt(2) <= T(2) && pt(3) >= S(3) && pt(3) <= T(3)
            cells.mVP_num{cur_cell} = cells.mVP_num{cur_cell} + 1;
            idx = cells.mVP_num{cur_cell};
            cells.mVP{cur_cell}(idx,:) = pt;
            cells.mVN{cur_cell}(idx,:) = mVN_orig(l,:);
          end
        end
      end
    end

    prog = (l/szmVPorig);
    waitbar(prog, h, sprintf('Computing number of points inside each cell (%5.1f%%)', prog*100));    
  end
end
delete(h);

cen_global = zeros(size(mVP_orig,1), 7);

% Compute number of non-empty cells and average number of points per cell
valid_cells = 0;
sum_pt_per_cell = 0;
for cur_cell=1:cells.number
  if cells.mVP_num{cur_cell} > 0
    sum_pt_per_cell = sum_pt_per_cell + cells.mVP_num{cur_cell};
    valid_cells = valid_cells + 1;
  end
end

[h, handles] = lshrbf_pou_gui_eval();
tot_cells = pou_div(1)*pou_div(2)*pou_div(3);
set(handles.text1, 'String', sprintf('Non-empty cells: %d', valid_cells));
set(handles.text2, 'String', sprintf('Average number of points per cell: %1.2f', sum_pt_per_cell / valid_cells));
ap = get(handles.axes2, 'Position');
set(handles.axes2, 'Position', [ap(1) ap(2) 0.01 ap(4)]);
ap = get(handles.axes3, 'Position');
set(handles.axes3, 'Position', [ap(1) ap(2) 0.01 ap(4)]);  
set(handles.text4, 'String', sprintf('Solving coefficients... (%5.1f%%)', 0));

tic;
% Save centers and coefficients to disk
coeff_filename = fullfile(dirout, fileout);
fid = fopen(coeff_filename, 'w');
fwrite(fid, num_cen, 'single'); % Maximum centers per cell (1 element)
fwrite(fid, cells.number, 'single'); % fprintf(fid, '%d\n', cells.number);
fwrite(fid, pou_div, 'single'); % pou division (3 elements)
fwrite(fid, pou_overlap, 'single'); % pou overlap (1 element)
fwrite(fid, min_pou, 'single'); % Bounding box min (3 elements)
fwrite(fid, max_pou, 'single'); % Bounding box max (3 elements)

% Assemble system
% For each cell...
cen_tot = 0;
cur_valid_cells = 0;
for cur_cell=1:cells.number
  
  if getappdata(h,'cancelled')
    delete(h);
    return;
  end  
  
  S = cells.S(cur_cell,:);
  T = cells.T(cur_cell,:);
  
  mVP_num = cells.mVP_num{cur_cell};
  
  num_con = 0;
  
  prog = (cur_cell/cells.number);
  set(handles.text4, 'String', sprintf('Solving coefficients... (%5.1f%%)   Cell %d/%d', prog*100, cur_cell, tot_cells));  
  ap = get(handles.axes2, 'Position');
  set(handles.axes2, 'Position', [ap(1) ap(2) prog*83 ap(4)]);  

  clear cen; clear con; clear cencon;
  cen = zeros(1, 7);
  cur_num_cen = num_cen;
  if cur_num_cen > mVP_num
    cur_num_cen = mVP_num;
  end
  
  clear mVP; clear mVN;
  cur_cen = 1;
  
  if cur_num_cen == 0 % Ignore cell if there are no centers in it
    cells.valid{cur_cell} = false;
    
    % Select cell's center and compute minimum distance to the point cloud
    cur_num_cen = 1;
    
    % First center is the point closest to the cell's center
    cell_center = S+((T-S)/2);
    
    [idx,dist] = dsearchn(mVP_orig, cell_center);
    mVP = mVP_orig(idx,:);
    mVN = mVN_orig(idx,:);    
    sig = sign(dot(mVN, (cell_center-mVP)));  
    
    mVN = cell_center - pou_center;
    if norm(mVN) > norm(max_pou - pou_center) / 2
      sig = 1;
      mVN = mVN / norm(mVN);
    end
    
    cen(1,:) = [cell_center mVN dist*sig];
  else
    mVP = cells.mVP{cur_cell};
    mVN = cells.mVN{cur_cell};
    
    cells.valid{cur_cell} = true;
    cur_valid_cells = cur_valid_cells + 1;
    
    % First center randomly selected
    %    idx = ceil(rand()*cur_num_cen);
    %    cen(1,:) = [mVP(idx,1) mVP(idx,2) mVP(idx,3) mVN(idx,1) mVN(idx,2) mVN(idx,3) 0];
    %    cen_global(num_cen_global,:) = cen(1,:); num_cen_global = num_cen_global + 1;
    
    % First center is the point closest to cell's center
    cell_center = S+((T-S)/2);
    idx = dsearchn(mVP, cell_center);
    cen(1,:) = [mVP(idx,1) mVP(idx,2) mVP(idx,3) mVN(idx,1) mVN(idx,2) mVN(idx,3) 0];
    cen_global(num_cen_global,:) = cen(1,:); num_cen_global = num_cen_global + 1;
       
    min_pt = min(mVP);
    max_pt = max(mVP);
    max_dist_cell = norm(max_pt - min_pt);
    
    best_dist = 0;
  end
  
  cen_tot = cen_tot + cur_num_cen;        
    
  num_con2 = 1;
  have_empty_region = false;
  num_extra_con = 0;
  while cur_cen <= cur_num_cen && num_con2 > 0 || have_empty_region    
    cputime_elapsed = toc;
    set(handles.text3, 'String', sprintf('Center %d/%d sampled from %d points in cell', cur_cen, cur_num_cen, mVP_num));
    set(handles.text5, 'String', sprintf('Selected centers: %d', cen_tot));    
    set(handles.text6, 'String', sprintf('Time elapsed: %02d:%02d:%02d', floor(cputime_elapsed/3600), round(mod(floor(cputime_elapsed/60),60)), round(mod(cputime_elapsed,60))));
    ap = get(handles.axes3, 'Position');
    set(handles.axes3, 'Position', [ap(1) ap(2) cur_cen/cur_num_cen*83 ap(4)]);    
    drawnow;

    if getappdata(h,'cancelled')
      delete(h);
      return;
    end        
   
    if cells.valid{cur_cell}
      % Constraints are con_percentage points which are not centers
      if num_extra_con <= 1
        tmp = cen(:,1:7)';
        tmp = tmp(:)';
        num_con = 0;
        con = zeros(size(mVP,1),7);
        cen_all = [mVP(:,1) mVP(:,2) mVP(:,3) mVN(:,1) mVN(:,2) mVN(:,3) zeros(size(mVP,1),1)];

        for i=1:size(mVP,1)
          if isempty(strfind(tmp, cen_all(i,1:7))) 
            num_con=num_con+1;
            con(num_con,1:7) = cen_all(i,1:7); 
          end
        end
        con_tmp = con;
        clear con;
        con = con_tmp(1:num_con,:);
        clear con_tmp;
        
        % con2 contains all points which are not centers
        num_con2 = 0;
        con2 = zeros(size(mVP,1),7);
        for i=1:size(mVP,1)
          if isempty(strfind(tmp, cen_all(i,1:7)))
            num_con2=num_con2+1;
            con2(num_con2,1:7) = cen_all(i,1:7);
          end
        end
        con_tmp = con2;
        clear con2;
        con2 = con_tmp(1:num_con2,:);
        clear con_tmp;
      end
      
      % After choosing all centers,
      % look for big empty regions that may require a constraint
      % to avoid false zero sheets
      
      % 'Discretize' centers
      dmcenters = zeros(dmsz);        
      cur_cel_cen = cen(1:cur_cen,1:3);
      clear iccc;
      iccc = ceil((cur_cel_cen - repmat(S,cur_cen,1)) ./ repmat((T-S), cur_cen, 1) .* repmat((dmsz-1), cur_cen, 1)) + repmat([1 1 1], cur_cen, 1);
      dmcenters(sub2ind(size(dmcenters),iccc(:,1), iccc(:,2), iccc(:,3))) = 1;
      
      while have_empty_region && use_least_squares && remove_false_sheets
        dm = dmcenters;
                
        % 'Discretize' constraints
        cur_cel_con = con(1:num_con,1:3);
        clear iccc;
        iccc = ceil((cur_cel_con - repmat(S,num_con,1)) ./ repmat((T-S), num_con, 1) .* repmat((dmsz-1), num_con, 1)) + repmat([1 1 1], num_con, 1);
        dm(sub2ind(size(dm),iccc(:,1), iccc(:,2), iccc(:,3))) = 1;

        dmd = bwdist(dm); % Compute distance map
        [dmd_max dmd_idx] = max(dmd(:));
        if dmd_max > dmsz(1)/4 % If maximum distance greater than threshold...
          % ...insert constraint at maximum distance point
          [sub_i, sub_j, sub_k] = ind2sub(dmsz, dmd_idx);
          ccc = [sub_i, sub_j, sub_k];
          ccc = (ccc-1) ./ (dmsz-1);
          ccc = (ccc .* (T-S)) + S;
          
          [idx,dist] = dsearchn(mVP_orig, ccc);
          mVP2 = mVP_orig(idx,:);
          mVN2 = mVN_orig(idx,:);
          sig = sign(dot(mVN2, (ccc-mVP2)));
                   
          tmp = [ccc mVN2 dist*sig];
          con_tst = con';
          if isempty(strfind(con_tst(:)', tmp))
            num_extra_con = num_extra_con + 1;
            num_con = num_con + 1;
            con(num_con,:) = tmp;
          else
            break;
          end
          clear con_tst;
          
          if num_extra_con > 1024 % Limit of constraints            
            have_empty_region = false;
          end
        else
          break;
        end
      end
    else
      num_con = 8+42;
      
      cell_indist = norm((T-S)/8);
      cell_pts = zeros(8, 3);
      cell_pts(1,1:3) = S; % Left bottom front corner
      cell_pts(2,1:3) = [T(1) S(2) S(3)]; % Right bottom front corner
      cell_pts(3,1:3) = [T(1) T(2) S(3)]; % Right top front corner
      cell_pts(4,1:3) = [S(1) T(2) S(3)]; % Left top front corner
      cell_pts(5,1:3) = T; % Right top back corner
      cell_pts(6,1:3) = [T(1) S(2) T(3)]; % Right bottom back corner
      cell_pts(7,1:3) = [S(1) S(2) T(3)]; % Left bottom back corner
      cell_pts(8,1:3) = [S(1) T(2) T(3)]; % Left top back corner
      
      for i=1:8 % First eight points are offset corners
        cell_indir = cell_center - cell_pts(i,1:3);
        cell_indir = cell_indir ./ norm(cell_indir);
        cell_pts(i,1:3) = cell_pts(i,1:3)+(cell_indir.*cell_indist);
        
        [idx,dist] = dsearchn(mVP_orig, cell_pts(i,1:3));
        mVP = mVP_orig(idx,:);
        mVN = mVN_orig(idx,:);
        sig = sign(dot(mVN, (cell_pts(i,1:3)-mVP)));
        con(i,:) = [cell_pts(i,1:3) mVN dist*sig];
      end
      
      i = 1;
      while i < 43 % Append contraints with random locations
        pts = S + rand(1,3).*(T-S);
        
        [idx,dist] = dsearchn(mVP_orig, pts);
        mVP = mVP_orig(idx,:);
        mVN = mVN_orig(idx,:);
        sig = sign(dot(mVN, (pts-mVP)));
        
        tmp = [pts mVN dist*sig];
        con_tst = con';
        if isempty(strfind(con_tst(:)', tmp))
          con(i+8,:) = tmp;
          i = i + 1;
        end
      end
    end
    
    % Centers + constraints
    num_cencon = cur_cen+num_con;
    cencon = zeros(num_cencon,7);
    
    cencon(1:cur_cen,:) = [cen(:,1) cen(:,2) cen(:,3) cen(:,4) cen(:,5) cen(:,6) cen(:,7)]; % Centers
    if num_con
      cencon(cur_cen+1:cur_cen+num_con,:) = [con(:,1), con(:,2), con(:,3), con(:,4), con(:,5), con(:,6) con(:,7)]; % Constraints
    end
    
    %% Assemble level set linear system
    if use_least_squares
      % Solve using least squares
      LSG = zeros((num_cencon*4)+4,(cur_cen*4)+4);
      LSF = zeros((num_cencon*4)+4,1);
      
      [I J] = meshgrid(1:cur_cen, 1:num_cencon);
      DIFF1 = cencon(J,1:3)-cen(I,1:3);
      clear I; clear J;
      D2_1 = reshape(DIFF1(:,1),num_cencon,cur_cen);
      D2_2 = reshape(DIFF1(:,2),num_cencon,cur_cen);
      D2_3 = reshape(DIFF1(:,3),num_cencon,cur_cen);
      clear DIFF1;
      R = sqrt(D2_1.*D2_1 + D2_2.*D2_2 + D2_3.*D2_3);
      
      EV = R.*R.*R; % HRBF r3
      LSG(1:4:num_cencon*4,1:4:cur_cen*4) = Inf;
      LSG(LSG==Inf) = EV;
      
      EVGX = (3.*R.*D2_1); % Gradient of HRBF
      LSG(1:4:num_cencon*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGX;
      LSG(2:4:num_cencon*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGX; clear EVGX;
      EVGY = (3.*R.*D2_2);
      LSG(1:4:num_cencon*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGY;
      LSG(3:4:num_cencon*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGY; clear EVGY;
      EVGZ = (3.*R.*D2_3);
      LSG(1:4:num_cencon*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGZ;
      LSG(4:4:num_cencon*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGZ; clear EVGZ;
      
      EV = ((3./R).*(2.*D2_1.*D2_1 + D2_2.*D2_2 + D2_3.*D2_3)); % Hessian_00 r3
      EV(isnan(EV))=0;
      LSG(2:4:num_cencon*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(D2_1.*D2_1 + 2.*D2_2.*D2_2 + D2_3.*D2_3)); % Hessian_11 r3
      EV(isnan(EV))=0;
      LSG(3:4:num_cencon*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(D2_1.*D2_1 + D2_2.*D2_2 + 2.*D2_3.*D2_3)); % Hessian_22 r3
      EV(isnan(EV))=0;
      LSG(4:4:num_cencon*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(D2_1.*D2_2)); % Hessian_01/10 r3
      EV(isnan(EV))=0;
      LSG(2:4:num_cencon*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(3:4:num_cencon*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(D2_1.*D2_3)); % Hessian_02/20 r3
      EV(isnan(EV))=0;
      LSG(2:4:num_cencon*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(4:4:num_cencon*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(D2_2.*D2_3)); % Hessian_12/21 r3
      EV(isnan(EV))=0;
      LSG(3:4:num_cencon*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(4:4:num_cencon*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      
      clear EV; clear R;
      clear D2_1; clear D2_2; clear D2_3;
      
      % Fill last 4 columns with polynomial term
      LSG(1:num_cencon*4,cur_cen*4+1:cur_cen*4+4) = repmat(eye(4),num_cencon,1);
      LSG(1:4:num_cencon*4,cur_cen*4+2) = cencon(1:num_cencon,1);
      LSG(1:4:num_cencon*4,cur_cen*4+3) = cencon(1:num_cencon,2);
      LSG(1:4:num_cencon*4,cur_cen*4+4) = cencon(1:num_cencon,3);
      
      LSG(num_cencon*4+1:num_cencon*4+4,1:cur_cen*4) = repmat(eye(4),1,cur_cen);
      LSG(num_cencon*4+2,1:4:cur_cen*4) = cen(1:cur_cen,1);
      LSG(num_cencon*4+3,1:4:cur_cen*4) = cen(1:cur_cen,2);
      LSG(num_cencon*4+4,1:4:cur_cen*4) = cen(1:cur_cen,3);
      
      % Fill F
      LSF(1:4:num_cencon*4) = cencon(1:num_cencon,7);      
      LSF(2:4:num_cencon*4) = cencon(1:num_cencon,4);
      LSF(3:4:num_cencon*4) = cencon(1:num_cencon,5);
      LSF(4:4:num_cencon*4) = cencon(1:num_cencon,6);
      
      LSA = LSG \ LSF;              
    else
      LSG = zeros((cur_cen*4)+4,(cur_cen*4)+4);
      LSF = zeros((cur_cen*4)+4,1);
      
      [I J] = meshgrid(1:cur_cen, 1:cur_cen);
      DIFF1 = cen(J,1:3)-cen(I,1:3);
      clear I; clear J;
      DIFF2(1:cur_cen,1:cur_cen,1) = reshape(DIFF1(:,1),cur_cen,cur_cen);
      DIFF2(1:cur_cen,1:cur_cen,2) = reshape(DIFF1(:,2),cur_cen,cur_cen);
      DIFF2(1:cur_cen,1:cur_cen,3) = reshape(DIFF1(:,3),cur_cen,cur_cen);
      clear DIFF1;
      R = sqrt(DIFF2(:,:,1).*DIFF2(:,:,1)+DIFF2(:,:,2).*DIFF2(:,:,2)+DIFF2(:,:,3).*DIFF2(:,:,3));
      
      EV = R.*R.*R; % HRBF r3
      LSG(1:4:cur_cen*4,1:4:cur_cen*4) = Inf;
      LSG(LSG==Inf) = EV;
      
      EVGX = (3.*R.*DIFF2(:,:,1)); % Gradient of HRBF r3
      EVGY = (3.*R.*DIFF2(:,:,2));
      EVGZ = (3.*R.*DIFF2(:,:,3));
      LSG(1:4:cur_cen*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGX;
      LSG(1:4:cur_cen*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGY;
      LSG(1:4:cur_cen*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EVGZ;
      LSG(2:4:cur_cen*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGX;
      LSG(3:4:cur_cen*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGY;
      LSG(4:4:cur_cen*4,1:4:cur_cen*4) = Inf; LSG(LSG==Inf) =  EVGZ;
      
      EV = ((3./R).*(2.*DIFF2(:,:,1).^2 + DIFF2(:,:,2).^2 + DIFF2(:,:,3).^2)); % Hessian_00 r3
      EV(isnan(EV))=0;
      LSG(2:4:cur_cen*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(DIFF2(:,:,1).^2 + 2.*DIFF2(:,:,2).^2 + DIFF2(:,:,3).^2)); % Hessian_11 r3
      EV(isnan(EV))=0;
      LSG(3:4:cur_cen*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(DIFF2(:,:,1).^2 + DIFF2(:,:,2).^2 + 2.*DIFF2(:,:,3).^2)); % Hessian_22 r3
      EV(isnan(EV))=0;
      LSG(4:4:cur_cen*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(DIFF2(:,:,1).*DIFF2(:,:,2))); % Hessian_01/10 r3 (12)
      EV(isnan(EV))=0;
      LSG(2:4:cur_cen*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(3:4:cur_cen*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(DIFF2(:,:,1).*DIFF2(:,:,3))); % Hessian_02/20 r3 (13)
      EV(isnan(EV))=0;
      LSG(2:4:cur_cen*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(4:4:cur_cen*4,2:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      EV = ((3./R).*(DIFF2(:,:,2).*DIFF2(:,:,3))); % Hessian_12/21 r3 (23)
      EV(isnan(EV))=0;
      LSG(3:4:cur_cen*4,4:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      LSG(4:4:cur_cen*4,3:4:cur_cen*4) = Inf; LSG(LSG==Inf) = -EV;
      clear EV; clear R; clear DIFF2;
      
      % Fill last 4 columns and 4 rows with polynomial term
      LSG(1:cur_cen*4,cur_cen*4+1:cur_cen*4+4) = repmat(eye(4),cur_cen,1);
      LSG(1:4:cur_cen*4,cur_cen*4+2) = cen(1:cur_cen,1);
      LSG(1:4:cur_cen*4,cur_cen*4+3) = cen(1:cur_cen,2);
      LSG(1:4:cur_cen*4,cur_cen*4+4) = cen(1:cur_cen,3);
      LSG(cur_cen*4+1:cur_cen*4+4,1:cur_cen*4) = LSG(1:cur_cen*4,cur_cen*4+1:cur_cen*4+4)';
      
      % Fill F
      LSF(1:4:cur_cen*4) = cen(1:cur_cen,7);
      LSF(2:4:cur_cen*4) = cen(1:cur_cen,4);
      LSF(3:4:cur_cen*4) = cen(1:cur_cen,5);
      LSF(4:4:cur_cen*4) = cen(1:cur_cen,6);
      
      LSA = LSG \ LSF;
    end
    clear LSG; clear LSF;
    
    if cur_cen < cur_num_cen
      % Evaluate system for each non-center position
      max_lsv = 0;
      max_lsv_j = 1;
      hessian = zeros(1,6);

      if use_random
        cur_cen = cur_cen - 1;
        num_cen_global = num_cen_global - 1;
        rand_idx = randperm(size(mVP,1));        
        for j=1:cur_num_cen
          cur_cen = cur_cen + 1;
          cen(cur_cen,:) = [mVP(rand_idx(j),:) mVN(rand_idx(j),:) 0];
          cen_global(num_cen_global,:) = cen(cur_cen,:); 
          num_cen_global = num_cen_global + 1;
        end          
      elseif use_filename_centers
        cur_cen = cur_cen - 1;
        num_cen_global = num_cen_global - 1;
        for j=1:num_loaded_centers
          cur_cen = cur_cen + 1;
          cen(cur_cen,:) = [loaded_centers(j,:) 0];
          cen_global(num_cen_global,:) = cen(cur_cen,:); 
          num_cen_global = num_cen_global + 1;
        end
      else
        %[tmpa, dist_list] = dsearchn(cen(:,1:3), con2(1:num_con2,1:3)); % r2009
        [~, dist_list] = knnsearch(cen(:,1:3), con2(1:num_con2,1:3)); % r2011+
        max_dist = max(dist_list);
        dist_factor = dist_list ./ max_dist;        
        
        for j=1:num_con2
          % Evaluation
          lsv = 0;
          glsv = zeros(1,3);
          con2j = con2(j,1:3);
          for ip=1:cur_cen
            diff = con2j-cen(ip,1:3);
            r = norm(diff);
            evg = (3*r)*diff;
            cC = LSA((ip-1)*4+1:(ip-1)*4+4)';            
            
            lsv = lsv + ((cC(1)*r*r*r) - (cC(2)*evg(1) + cC(3)*evg(2) + cC(4)*evg(3)));
            glsv = glsv + ((cC(1)*evg) - cC(2:4) * 3 / r * (r * r * eye3 + diff'*diff));
          end

          % Polynomial term
          cC = LSA(cur_cen*4+1:cur_cen*4+4)';
          lsv = lsv + cC(1) + cC(2)*con2j(1) + cC(3)*con2j(2) + cC(4)*con2j(3);
          glsv = glsv + cC(2:4);
          
          error = norm([lsv*lsv_scale (glsv-con2(j,4:6))*greedy_options(2)]) * dist_factor(j);
          
          if error > max_lsv
            max_lsv_j = j;
            max_lsv = error;
          end
        end
        
        % Add constraint with maximum distance from currently evaluated HRBF
        if num_con2 > 0
          cur_cen = cur_cen + 1;
          cen(cur_cen,:) = con2(max_lsv_j,:);
          cen_global(num_cen_global,:) = cen(cur_cen,:); num_cen_global = num_cen_global + 1;
        end
      end
    else
      if have_empty_region
        break;
      end
      have_empty_region = true;
    end   
        
    if ~cells.valid{cur_cell}
      break
    end
  end
     
  fwrite(fid, cur_cell, 'single');
  fwrite(fid, cells.valid{cur_cell}, 'single');
  fwrite(fid, cur_num_cen, 'single');
  for i=1:cur_num_cen
    fwrite(fid, cen(i,1:6), 'single');
    fwrite(fid, LSA(((i-1)*4)+1:((i-1)*4)+4), 'single');
  end
  % Polynomial coefficients
  fwrite(fid, LSA((i*4)+1:(i*4)+4), 'single');
   
  cells.cen{cur_cell} = cen;
  cells.LSA{cur_cell} = LSA;
end

fclose(fid); % Close coefficient file

solve_cputime_elapsed = cputime_elapsed;

% Evaluate
tic;
if eval_options(1) || eval_options(2)
  LSV = zeros(max(cells.dT,[],1));
  AcW = zeros(size(LSV));

  % Start parallel processing pool
  %if matlabpool('size') > 0
  %  matlabpool close;
  %end
  %matlabpool open;

  % For each cell...
  set(handles.text4, 'String', sprintf('Evaluating domain... (%5.1f%%)', 0));
  ap = get(handles.axes2, 'Position');
  set(handles.axes2, 'Position', [ap(1) ap(2) 0.01 ap(4)]);    
  ap = get(handles.axes3, 'Position');
  set(handles.axes3, 'Position', [ap(1) ap(2) 0.01 ap(4)]);    
  drawnow;  

  cur_valid_cells = 0;
  for cur_cell=1:cells.number
    S = cells.S(cur_cell,:) + half_cell;
    T = cells.T(cur_cell,:) - half_cell;
    dS = cells.dS(cur_cell,:);
    dT = cells.dT(cur_cell,:);

    step = (T-S)./(dT-dS);

    clear I; clear J; clear K;
    [I J K] = meshgrid(S(1):step(1):T(1)+(step(1)/2), S(2):step(2):T(2)+(step(2)/2), S(3):step(3):T(3)+(step(3)/2));

    I = permute(I, [2 1 3]);
    J = permute(J, [2 1 3]);
    K = permute(K, [2 1 3]);

    S = cells.S(cur_cell,:);
    T = cells.T(cur_cell,:);
    DFc =         4 * ( (I-S(1)).*(T(1)-I) ./ ((T(1)-S(1)).^2) );
    DFc = DFc .* (4 * ( (J-S(2)).*(T(2)-J) ./ ((T(2)-S(2)).^2) ));
    DFc = DFc .* (4 * ( (K-S(3)).*(T(3)-K) ./ ((T(3)-S(3)).^2) ));
    DFc = 1 - DFc;
    CFc = -6*DFc.^5 + 15*DFc.^4 - 10*DFc.^3 + 1;

    cur_valid_cells = cur_valid_cells + 1;

    prog = (cur_cell/cells.number);
    set(handles.text4, 'String', sprintf('Evaluating domain... (%5.1f%%)   Cell %d/%d', prog*100, cur_cell, cells.number));  
    ap = get(handles.axes2, 'Position');
    set(handles.axes2, 'Position', [ap(1) ap(2) prog*83 ap(4)]);
    cputime_elapsed = toc;
    set(handles.text6, 'String', sprintf('Time elapsed: %02d:%02d:%02d', floor(cputime_elapsed/3600), round(mod(floor(cputime_elapsed/60),60)), round(mod(cputime_elapsed,60))));
    drawnow;

    % Evaluate current [I J K] cell, weighted by CFc
    LSA = cells.LSA{cur_cell};

    [M_loc N_loc O_loc] = size(CFc);

    LSX(:,:,:,1)=I;
    LSX(:,:,:,2)=J;
    LSX(:,:,:,3)=K;

    LSVloc = zeros(M_loc, N_loc, O_loc);

    cen = cells.cen{cur_cell};
    num_cen = size(cen,1);

    %parfor_progress(num_cen);
    %parfor i=1:num_cen
    for i=1:num_cen
      set(handles.text3, 'String', sprintf('Center %d/%d', i, num_cen));    
      ap = get(handles.axes3, 'Position');
      set(handles.axes3, 'Position', [ap(1) ap(2) i/num_cen*83 ap(4)]);    
      drawnow;  

      ce = cen;
      lsa = LSA;
      lsx = LSX;
      LSYX = repmat(ce(i,1),[M_loc N_loc O_loc]);
      LSYY = repmat(ce(i,2),[M_loc N_loc O_loc]);
      LSYZ = repmat(ce(i,3),[M_loc N_loc O_loc]);
      DIFFX = lsx(:,:,:,1)-LSYX;
      DIFFY = lsx(:,:,:,2)-LSYY;
      DIFFZ = lsx(:,:,:,3)-LSYZ;
      R = sqrt((DIFFX.*DIFFX + DIFFY.*DIFFY + DIFFZ.*DIFFZ));

      EV = R.*R.*R;
      EVGX = (3.*R).*DIFFX;
      EVGY = (3.*R).*DIFFY;
      EVGZ = (3.*R).*DIFFZ;

      tmp = (lsa((i-1)*4+1) .* EV) - ((lsa((i-1)*4+2).*EVGX) + (lsa((i-1)*4+3).*EVGY) + (lsa((i-1)*4+4).*EVGZ));
      LSVloc = LSVloc + tmp;
      %parfor_progress;
    end

    % Polynomial term
    LSVloc = LSVloc + LSA(num_cen*4+1) + LSA(num_cen*4+2).*LSX(:,:,:,1) + LSA(num_cen*4+3).*LSX(:,:,:,2) + LSA(num_cen*4+4).*LSX(:,:,:,3);

    %parfor_progress(0);

    clear EV; clear EVGX; clear EVGY; clear EVGZ; clear LSX; clear LSY; clear DIFF; clear R;

    % Add to global domain
    AcW(dS(1):dT(1), dS(2):dT(2), dS(3):dT(3)) = AcW(dS(1):dT(1), dS(2):dT(2), dS(3):dT(3)) + CFc;
    LSV(dS(1):dT(1), dS(2):dT(2), dS(3):dT(3)) = LSV(dS(1):dT(1), dS(2):dT(2), dS(3):dT(3)) + (LSVloc .* CFc);
  end

  %matlabpool close;

  % Normalize by sum of weights of overlapping cells
  LSV = LSV ./ AcW;
end

delete(h);

uiwait(msgbox(sprintf('Implicit reconstruction done in %02d:%02d:%02d with %d centers.', floor(solve_cputime_elapsed/3600), round(mod(floor(solve_cputime_elapsed/60),60)), round(mod(solve_cputime_elapsed,60)), cen_tot), 'Done!'));

close all;

if eval_options(1)
  figure('Name','Scalar Field','NumberTitle','off')
  grid on;
  set(gcf, 'Color', 'w');
  vol3d('cdata',LSV,'texture','3D');
  axis([0, size(LSV,2), 0, size(LSV,1), 0, size(LSV,3)]);
  daspect([1 1 1]);
  alphamap('rampup'); alphamap(.5 .* alphamap);
  view(-45,45);
end

% Show isosurface
if eval_options(2)
  figure('Name','Zero-level Isosurface','NumberTitle','off')
  set(gcf, 'Color', 'w');
  grid on;
  LSV2 = LSV;
  [f,v]=isosurface(LSV2,0);
  if eval_options(5)
    vertface2obj(v, f, fullfile(eval_dirout, eval_fileout));
  end
  p = patch(isosurface(LSV2,0));
  isonormals(LSV2,p);
  set(p,'FaceColor',[.9 .9 .9],'EdgeColor','none');
  daspect([1 1 1]);
  cl1 = camlight(-20,  55,'infinite'); set(cl1, 'Color', [1    0.65 0.15]);
  cl3 = camlight( 135,-20,'infinite'); set(cl3, 'Color', [1    0.55 0.15]);
  cl2 = camlight( 20, -55,'infinite'); set(cl2, 'Color', [0.15 0.65 1   ]);
  cl4 = camlight(-135, 20,'infinite'); set(cl4, 'Color', [0.15 0.55 1   ]);

  material dull;
  lighting phong;
  alpha(1);
  
  [M_tot N_tot O_tot] = size(LSV2);  
  
  axis([1 N_tot 1 M_tot 1 O_tot]); axis off;
  view(-45,45); 
end

if eval_options(3)
  figure('Name','Point Cloud and Selected Centers','NumberTitle','off')
  set(gcf, 'Color', 'w');
  grid on;
  axis off;
  hp=plot3(mVP_orig(:,2), mVP_orig(:,1), mVP_orig(:,3), 'b.');
  set(hp, 'MarkerSize', 1);  
  hold on;
  pt = cen_global(1:num_cen_global-1,1:3);
  scatter3(pt(:,2), pt(:,1), pt(:,3), 'ro');
  quiver3(pt(:,2), pt(:,1), pt(:,3), cen_global(1:num_cen_global-1,5), cen_global(1:num_cen_global-1,4), cen_global(1:num_cen_global-1,6), 0.3, 'r');
  daspect([1 1 1]);
  hold off;
  view(-45,45);
end  

end

